/*
 * Decompression utility for AMI BIOSes.
 *
 * Copyright (C) 2009-2010 coresystems GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "compat.h"

#if (defined(__APPLE__) || defined(__DJGPP__) || defined(__MINGW32__))
void *memmem(const void *haystack, size_t haystacklen, const void *needle,
	     size_t needlelen)
{
	char *searchpointer = (char *) haystack;
	char *patternpointer = (char *) needle;
	char *endofsearch = searchpointer + haystacklen - needlelen;

	if (!(haystack && needle && haystacklen && needlelen))
		return NULL;

	while (searchpointer <= endofsearch) {
		if (*searchpointer == *patternpointer)
			if (memcmp
			    (searchpointer, patternpointer,
			     needlelen) == 0)
				return searchpointer;

		searchpointer++;
	}

	return NULL;
}

size_t strnlen(const char *s, size_t maxlen)
{
	const char *end = memchr(s, '\0', maxlen);
	return end ? (size_t) (end - s) : maxlen;
}

char *strndup(const char *s, size_t n)
{
	size_t len = strnlen(s, n);
	char *new = malloc(len + 1);

	if (new == NULL)
		return NULL;

	new[len] = '\0';
	return memcpy(new, s, len);
}
#endif

#if (defined(__DJGPP__) || defined(__MINGW32__))
// DJGPP & MINGW needs mmap emulation, this one works only for files using buffers in RAM
static MMAP_ENTRY mmap_save_buffer[MMAP_ENTRIES]; // array for holding info about mapped areas
static int mmap_save_index=0;          // index of mapped areas

// create file to memory mapping (create virtual buffer)
void *mmap(void *addr, size_t len, int prot, int flags, int fd, off_t offset)
{
  void *buffer=NULL;                   // buffer for mapped (parts of) file
#ifdef DEBUG_MMAP
  printf("mmap(%ph, %d, %Xh, %Xh, %d, %ld)\n", addr, len, prot, flags, fd, offset);
#endif
  if (addr!=NULL)                      // check addr
    {
    printf("mmap() doesn't support nonzero address!\n");
    return((void *)-1);                // return MAP_FAILED
    }
  if (fd<0)                            // check fd
    {
    printf("mmap() got invalid file descriptor!\n");
    return((void *)-1);                // return MAP_FAILED
    }
  if ((buffer=malloc(len))==NULL)      // allocate memory for buffer
    {
    printf("mmap() failed to allocate %ld Bytes!\n", (long)len);
    return((void *)-1);                // return MAP_FAILED
    }
  if (prot==PROT_READ)                 // for read mode
    {
    lseek(fd, offset, SEEK_SET);       // set file offset
    read(fd, buffer, len);             // read part of the file to buffer
    }
  else if ((prot&PROT_WRITE)!=0)       // for write mode
    memset(buffer, 0, len);
  else
    {
    printf("mmap() flags unsupported!\n");
    return((void *)-1);                // return MAP_FAILED
    }
  if (mmap_save_index>=MMAP_ENTRIES)   // check mmap_save_index
    {
    printf("mmap() call buffer overflow!\n");
    return((void *)-1);                // return MAP_FAILED
    }
  else
    {                                  // save mapped area params to array
    mmap_save_buffer[mmap_save_index].addr=buffer;
    mmap_save_buffer[mmap_save_index].fd=fd;
    mmap_save_buffer[mmap_save_index].prot=prot;
    mmap_save_buffer[mmap_save_index].offset=offset;
    mmap_save_index++;                 // update array index
    }
#ifdef DEBUG_MMAP
  printf("mmap returned allocated buffer at: %ph\n", buffer);
#endif
  return(buffer);                      // return pointer to allocated buffer
}

// remove file to memory mapping (free virtual buffer and close file)
int munmap(void *addr, size_t len)     // addr is virtual address returned by mmap()
{
#ifdef DEBUG_MMAP
  printf("munmap(%ph, %u)\n", addr, len);
#endif
  if ((addr==NULL) || (addr==(void *)-1)) // check addr for invalid value
    return(-1);
  if (mmap_save_index<=0)              // check mmap_save_index
    {
    printf("mmap() call buffer underflow!\n");
    return(-1);
    }
  mmap_save_index--;
  if (mmap_save_buffer[mmap_save_index].addr!=addr) // check addr if match with stored entry
    {
    printf("mmap()/munmap() call unpaired!\n");
    return(-1);
    }
  if ((mmap_save_buffer[mmap_save_index].prot&PROT_WRITE)!=0) // for write mode
    {
    lseek(mmap_save_buffer[mmap_save_index].fd, mmap_save_buffer[mmap_save_index].offset, SEEK_SET); // set file offset
    write(mmap_save_buffer[mmap_save_index].fd, addr, len); // save buffer to file
    close(mmap_save_buffer[mmap_save_index].fd); // close file
    }
  if (mmap_save_buffer[mmap_save_index].prot==PROT_READ) // for read mode
    close(mmap_save_buffer[mmap_save_index].fd); // close file
  free(addr);                          // free previously allocated memory
  return(0);
}

#endif
